from importlib import resources

from rich.traceback import install

from .exceptions import InstaterError
from .main import run_tasks

install()

VERSION = resources.read_text("instater", "VERSION").strip()

__all__ = [
    "InstaterError",
    "VERSION",
    "run_tasks",
]



import shutil
import tempfile
from typing import List, Set, Union

from instater.exceptions import InstaterError

from .. import util
from ..context import Context
from . import Task


def _not_installed(package: str) -> bool:
    result_package = util.shell(["pacman", "-Qi", package], valid_return_codes=(0, 1))
    result_group = util.shell(["pacman", "-Qg", package], valid_return_codes=(0, 1))
    return result_package.return_code != 0 and result_group.return_code != 0


class Pacman(Task):
    def __init__(self, *, packages: Union[str, List[str]], aur: util.Bool = False, become: str = None, **kwargs):
        super().__init__(**kwargs)

        if isinstance(packages, str):
            packages = [packages]

        self.packages = packages
        self.aur = util.boolean(aur)
        self.become = become

        if self.become and not self.aur:
            raise InstaterError("Can only specify 'become' when using 'aur'")

    # TODO: this currently requires root to run properly (to delete the cloned directory created by another user)
    def _makepkg_install_package(self, package: str):
        with tempfile.TemporaryDirectory() as tmpdir:
            if self.become:
                util.shell(["chown", self.become, tmpdir])

            clone_dir = tmpdir + "/" + package

            util.shell(
                ["git", "clone", "--depth", "1", f"https://aur.archlinux.org/{package}.git", clone_dir],
                become=self.become,
            )

            util.shell(
                ["makepkg", "--syncdeps", "--install", "--noconfirm", "--needed"],
                directory=clone_dir,
                become=self.become,
            )

    def _makepkg_install(self, packages: List[str]):
        for package in packages:
            self._makepkg_install_package(package)

    def _yay_install(self, packages: List[str]):
        util.shell(["yay", "-Sy", "--noconfirm", "--needed", "--cleanafter", *packages])

    def _pacman_install(self, packages: List[str]):
        util.shell(["pacman", "-Sy", "--noconfirm", "--noprogressbar", "--needed", *packages])

    def _install(self, packages: List[str]):
        if self.aur:
            if shutil.which("yay"):
                self._yay_install(packages)
            else:
                self._makepkg_install(packages)
        else:
            self._pacman_install(packages)

    def run_action(self, context: Context) -> bool:
        not_installed = list(filter(_not_installed, self.packages))

        if not not_installed:
            return False

        if not context.dry_run:
            self._install(not_installed)

        return True


class Aur(Pacman):
    def __init__(self, **kwargs):
        kwargs["aur"] = True
        super().__init__(**kwargs)


def get_explicitly_installed_packages() -> Set[str]:
    output = util.shell(["pacman", "-Qe"])
    return set(line.split()[0] for line in output.stdout.splitlines() if line)


def get_package_or_group_packages(package: str) -> Set[str]:
    group_output = util.shell(["pacman", "-Qg", package], valid_return_codes=(0, 1))
    if group_output.return_code == 1:
        return {package}
    else:
        return set(line.split()[1] for line in group_output.stdout.splitlines())



from typing import Iterable, List, Optional, Union

from .. import util
from ..context import Context
from . import Task


def _user_exists(user: str) -> bool:
    result = util.shell(["getent", "passwd", user], valid_return_codes=(0, 2))
    return result.return_code == 0


def _create_user(user: str, system: bool, create_home: bool, password: Optional[str]):
    command = ["useradd", user]

    if system:
        command.append("--system")

    if create_home:
        command.append("--create-home")

    if password:
        command.append("--password")
        command.append(password)

    util.shell(command)


def _list_groups(user: str) -> List[str]:
    result = util.shell(["groups", user])
    return result.stdout.split(" ")


def _add_groups(user: str, groups: Iterable[str]):
    group_str = ",".join(groups)
    util.shell(["usermod", "-a", "-G", group_str, user])


def _get_shell(user: str) -> str:
    result = util.shell(["getent", "passwd", user])
    return result.stdout.split(":")[-1]


def _set_shell(user: str, shell: str):
    util.shell(["usermod", "-s", shell, user])


class User(Task):
    def __init__(
        self,
        user: str,
        system: util.Bool = None,
        create_home: util.Bool = None,
        password: str = None,
        shell: str = None,
        groups: Union[str, List[str]] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        if isinstance(groups, str):
            groups = [groups]

        self.user = user
        self.system = util.boolean(system)
        self.create_home = util.boolean(create_home)
        self.password = password
        self.shell = shell
        self.groups = groups or []

    def run_action(self, context: Context) -> bool:
        updated = False

        user_exists = _user_exists(self.user)
        if not user_exists:
            if not context.dry_run:
                _create_user(self.user, self.system, self.create_home, self.password)
            updated = True
            missing_groups: Iterable[str] = self.groups
        else:
            all_groups = _list_groups(self.user)
            missing_groups = set(self.groups) - set(all_groups)

        if missing_groups:
            if not context.dry_run:
                _add_groups(self.user, self.groups)
            updated = True

        if self.shell is not None and self.shell != _get_shell(self.user):
            _set_shell(self.user, self.shell)
            updated = True

        # TODO: detect if password needs to change?

        return updated



import shlex
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Union
from urllib.request import urlopen

from .. import util
from ..context import Context
from ..exceptions import InstaterError
from . import Task


def _read(path: Path) -> bytes:
    with path.open("rb") as file:
        return file.read()


class Copy(Task):
    def __init__(
        self,
        *,
        src: str = None,
        content: str = None,
        url: str = None,
        dest: str,
        owner: str = None,
        group: str = None,
        mode: Union[str, int] = None,
        is_template: util.Bool = False,
        validate: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if not util.single_truthy(src, content, url):
            raise InstaterError("Must provide exactly one source of data to copy")

        if isinstance(mode, str):
            mode = int(mode, 8)

        self.src = Path(src) if src else None
        self.content = content
        self.url = url
        self.dest = Path(dest)
        self.owner = owner
        self.group = group
        self.mode = mode
        self.is_template = util.boolean(is_template)
        self.validate = validate

    def _update_metadata(self, file: Path, context: Context) -> bool:
        return util.update_file_metadata(file, self.owner, self.group, self.mode, context.dry_run)

    def _validate(self, path: Path):
        if not self.validate:
            return

        validate_command = shlex.split(self.validate % path)
        # util.shell will raise an error on exit codes > 0
        util.shell(validate_command)

    def _update_file_direct(self, src: Path, dest: Path, context: Context) -> bool:
        updated = False

        self._validate(src)

        if not dest.exists():
            if not context.dry_run:
                dest.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(src, dest)
            updated = True
        elif not src.samefile(dest) and _read(src) != _read(dest):
            if not context.dry_run:
                shutil.copy(src, dest)
            updated = True

        updated |= self._update_metadata(dest, context)
        return updated

    def _update_file_content(self, content: str, dest: Path, context: Context) -> bool:
        updated = False

        if content is not None and not content.endswith("\n"):
            content += "\n"

        if self.validate:
            with NamedTemporaryFile() as temp_file:
                temp_file.write(content.encode("utf-8"))
                temp_file.flush()
                self._validate(Path(temp_file.name))

        if not dest.exists():
            if not context.dry_run:
                dest.parent.mkdir(parents=True, exist_ok=True)
                with dest.open("w") as f:
                    f.write(content)
            updated = True
        elif content.encode("utf-8") != _read(dest):
            if not context.dry_run:
                with dest.open("w") as f:
                    f.write(content)
            updated = True

        updated |= self._update_metadata(dest, context)
        return updated

    def _update_file_template(self, src: Path, dest: Path, context: Context) -> bool:
        with src.open() as f:
            content = context.jinja_string(f.read())

        return self._update_file_content(content, dest, context)

    def _update_file(self, src: Path, dest: Path, context: Context):
        if self.is_template:
            return self._update_file_template(src, dest, context)
        else:
            return self._update_file_direct(src, dest, context)

    def _update_dir(self, context: Context) -> bool:
        updated = False
        src: Path = self.src  # type: ignore

        for path in src.glob("**/*"):
            if path.is_file():
                dest_path = self.dest / path.relative_to(src)
                if self._update_file(path, dest_path, context):
                    updated = True

        return updated

    def run_action(self, context: Context) -> bool:
        src = self.src
        if src and not src.is_absolute():
            src = context.root_directory / src

        content = self.content
        if self.url:
            content = urlopen(self.url).read().decode("utf-8")

        dest = self.dest
        if not self.dest.is_absolute():
            dest = context.root_directory / dest

        if src and not src.exists():
            raise InstaterError(f"Source to copy does not exist: {src}")

        if content is not None or (src and src.is_file()):
            if dest.exists() and not dest.is_file():
                raise InstaterError(f"Destination is a directory, expected file: {dest}")

            if src:
                return self._update_file(src, dest, context)
            else:
                content_str: str = content  # type: ignore
                if self.is_template:
                    content_str = context.jinja_string(content_str)

                return self._update_file_content(content_str, dest, context)
        else:
            if dest.exists() and not dest.is_dir():
                raise InstaterError(f"Destination is a file, expected directory: {dest}")

            return self._update_dir(context)


class Template(Copy):
    def __init__(self, **kwargs):
        kwargs.setdefault("is_template", True)
        super().__init__(**kwargs)



import json
from argparse import ArgumentParser

from rich.console import Console

from instater import VERSION, InstaterError, run_tasks


def _parse_variables(vars: str) -> dict:
    if vars is None:
        return {}

    try:
        parsed = json.loads(vars)
        if not isinstance(parsed, dict):
            raise InstaterError(f"JSON variables from --vars must be a dictionary, found '{type(parsed)}'")
        return parsed
    except json.JSONDecodeError:
        pass

    split = vars.replace(";", " ").split(" ")
    parsed = {}
    for item in split:
        item = item.strip()
        if not item:
            continue

        if "=" not in item:
            raise InstaterError(f"Invalid argument to --vars: '{item}' (missing '=')")
        key, value = item.split("=")
        parsed[key] = value

    return parsed


def main():
    parser = ArgumentParser(description="An easy solution for system/dotfile configuration")
    parser.add_argument("--setup-file", default="setup.yml", help="The setup file to execute")
    parser.add_argument("--tags", nargs="*", help="Run only a subset of tasks by their tag")
    parser.add_argument("--vars", help="Variables to override prompts or variable files")
    parser.add_argument(
        "--skip-tasks",
        action="store_true",
        help="Do not actually run tasks (useful for just checking for manually installed pacman packages)",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Display operations that would be performed without actually running them",
    )
    parser.add_argument("--version", action="store_true", help="Display the version of instater")

    args = parser.parse_args()

    if args.version:
        Console().print(f"Instater {VERSION}")
        return

    tags = args.tags

    try:
        variables = _parse_variables(args.vars)
        run_tasks(args.setup_file, variables, tags, args.dry_run, skip_tasks=args.skip_tasks)
    except InstaterError as e:
        console = Console()
        console.print(e, style="red")
        console.print("Exiting...", style="red bold")



import os
from pathlib import Path
from typing import Union

from instater.exceptions import InstaterError

from .. import util
from ..context import Context
from . import Task


class File(Task):
    def __init__(
        self,
        *,
        path: str,
        target: str = None,
        owner: str = None,
        group: str = None,
        mode: Union[str, int] = None,
        directory: util.Bool = False,
        symlink: util.Bool = False,
        hard_link: util.Bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if isinstance(mode, str):
            mode = int(mode, 8)

        if not util.single_truthy(directory, symlink, hard_link, allow_zero=True):
            raise InstaterError("Must only provide ony of directory, symlink, or hard_link")

        if target and (not symlink and not hard_link):
            raise InstaterError("Must provide a target with symlink/hard_link")

        self.path = Path(path)
        self.target = Path(target) if target is not None else None
        self.owner = owner
        self.group = group
        self.mode = mode
        self.directory = util.boolean(directory)
        self.symlink = util.boolean(symlink)
        self.hard_link = util.boolean(hard_link)

        if target is not None and not self.symlink and not self.hard_link:
            raise InstaterError("Argument `target` may only be used with symlink or hard_link files")

    def _create_file(self):
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.touch()

    def _create_directory(self):
        self.path.mkdir(parents=True, exist_ok=True)

    def _create_symlink(self):
        target: Path = self.target  # type: ignore
        os.symlink(target, self.path)

    def _create_hard_link(self):
        target: Path = self.target  # type: ignore
        os.link(target, self.path)

    def run_action(self, context: Context):
        updated = False

        if not self.path.exists():
            if not context.dry_run:
                if self.directory:
                    self._create_directory()
                elif self.symlink:
                    self._create_symlink()
                elif self.hard_link:
                    self._create_hard_link()
                else:
                    self._create_file()
            updated = True
        elif self.symlink:
            if not self.path.is_symlink():
                raise InstaterError(f"Path exists but is not a symlink: {self.path}")
        elif self.directory:
            if not self.path.is_dir():
                raise InstaterError(f"Path exists but is not a directory: {self.path}")
        elif self.hard_link:
            pass
        else:
            if not self.path.is_file():
                raise InstaterError(f"Path exists but is not a file: {self.path}")

        updated |= util.update_file_metadata(self.path, self.owner, self.group, self.mode, context.dry_run)

        return updated


class Directory(File):
    def __init__(self, **kwargs):
        kwargs["directory"] = True
        super().__init__(**kwargs)


class Symlink(File):
    def __init__(self, **kwargs):
        kwargs["symlink"] = True
        super().__init__(**kwargs)


class HardLink(File):
    def __init__(self, **kwargs):
        kwargs["hard_link"] = True
        super().__init__(**kwargs)



import os.path
import time
import typing
from collections import Counter
from pathlib import Path
from typing import Iterable

from jinja2 import Environment, FileSystemLoader
from rich.console import Console

from . import util


def _filename(path: str) -> str:
    return os.path.basename(path).rsplit(".", 1)[0]


def _jinja_environment(root_directory: Path) -> Environment:
    env = Environment(loader=FileSystemLoader(root_directory))
    env.filters["password_hash"] = util.password_hash
    env.filters["filename"] = _filename
    env.filters["bool"] = bool
    return env


class Context:
    def __init__(self, root_directory: Path, extra_vars: dict, tags: Iterable[str], dry_run: bool = False):
        self.root_directory = root_directory
        self.tags = set(tags)
        self.dry_run = dry_run

        extra_vars["instater_dir"] = str(root_directory.absolute())
        self.variables = extra_vars

        self.jinja_env = _jinja_environment(root_directory)
        self.tasks: list = []
        self.statuses: typing.Counter[str] = Counter()

        self.start = time.time()

        self.console = Console()
        self.print = self.console.print

    def jinja_object(self, template: object, extra_vars: dict = None) -> object:
        if isinstance(template, str):
            return self.jinja_string(template, extra_vars)
        elif isinstance(template, list):
            return [self.jinja_object(item, extra_vars) for item in template]
        else:
            return template

    def jinja_string(self, template: str, extra_vars: dict = None) -> str:
        if isinstance(template, str):
            vars = self.variables
            if extra_vars:
                vars = {**vars, **extra_vars}
            return self.jinja_env.from_string(template).render(vars)
        else:
            return template

    def jinja_file(self, template_path: str, extra_vars: dict = None) -> str:
        vars = self.variables
        if extra_vars:
            vars = {**vars, **extra_vars}
        return self.jinja_env.get_template(template_path).render(vars)

    def duration(self, start: float = None):
        if start is None:
            start = self.start

        duration = round(time.time() - start, 3)
        return f"[white]({duration}s)[/white]"

    def print_summary(self):
        skipped = self.statuses["skipped"]
        changed = self.statuses["changed"]

        self.print(f"Summary {self.duration()}:", style="bold")
        self.print(f"  skipped: {skipped}", style="blue")
        # if this used style="yellow", the integer count would be turned blue by rich
        self.print(f"  [yellow]changed: {changed}[/yellow]")



from .. import util
from ..context import Context
from . import Task


class Service(Task):
    def __init__(self, service: str, started: util.Bool = False, enabled: util.Bool = False, **kwargs):
        super().__init__(**kwargs)

        self.service = service
        self.started = util.boolean(started)
        self.enabled = util.boolean(enabled)

    def _is_started(self) -> bool:
        result = util.shell(["systemctl", "is-active", self.service], valid_return_codes=(0, 3))
        return result.stdout == "active"

    def _is_enabled(self) -> bool:
        result = util.shell(["systemctl", "is-enabled", self.service], valid_return_codes=(0, 1))
        return result.stdout == "enabled"

    def run_action(self, context: Context) -> bool:
        updated = False

        if self.started and not self._is_started():
            if not context.dry_run:
                util.shell(["systemctl", "start", self.service])
            updated = True

        if self.enabled and not self._is_enabled():
            if not context.dry_run:
                util.shell(["systemctl", "enable", self.service])
            updated = True
        elif not self.enabled and self._is_enabled():
            if not context.dry_run:
                util.shell(["systemctl", "disable", self.service])
            updated = True

        return updated



from .. import util
from ..context import Context
from . import Task


class Group(Task):
    def __init__(self, group: str, **kwargs):
        super().__init__(**kwargs)

        self.group = group

    def _group_exists(self) -> bool:
        result = util.shell(["getent", "group", self.group], valid_return_codes=(0, 2))
        return result.return_code == 0

    def run_action(self, context: Context) -> bool:
        if self._group_exists():
            return False

        if not context.dry_run:
            util.shell(["groupadd", self.group])

        return True



import getpass
import shutil
from glob import glob
from pathlib import Path
from typing import Iterable, List, Union

import yaml  # type: ignore

from . import util
from .context import Context
from .exceptions import InstaterError
from .tasks import TASKS, pacman


def _print_start(context: Context, setup_file: Path):
    context.print(f"Beginning instater execution from {setup_file.absolute()}", style="green bold")
    if context.tags:
        colored_tags = ", ".join(f"[bold]{tag}[/bold]" for tag in context.tags)
        context.print("Only executing specified tags:", colored_tags, style="blue")

    if context.variables:
        formatted_vars = " ".join(f"{key}={value!r}" for key, value in context.variables.items())
        context.print("Overridden variables:", formatted_vars, style="blue")

    print()


def _get_raw_input(prompt: str, private: bool) -> str:
    if private:
        return getpass.getpass(prompt)
    else:
        return input(prompt)


def _get_input(prompt: str, private: bool, allow_empty: bool) -> str:
    while True:
        raw_input = _get_raw_input(prompt, private)
        if allow_empty or raw_input:
            return raw_input


def _do_prompt(prompt: str, private, confirm, allow_empty) -> str:
    prompt += ": "
    while True:
        input = _get_input(prompt, private, allow_empty)
        if not confirm:
            return input

        confirm_input = _get_input("Confirm: ", private, allow_empty)
        if input == confirm_input:
            return input
        print("Inputs did not match")


def _prompt_variables(prompt_vars, context: Context):
    if not prompt_vars:
        return

    if not isinstance(prompt_vars, list):
        prompt_vars = [prompt_vars]

    for prompt_var in prompt_vars:
        if isinstance(prompt_var, str):
            prompt_var = {"name": prompt_var}

        name = prompt_var.get("name")
        if not name:
            raise InstaterError("Missing 'name' in vars_prompt")

        # skip variables that were defined in the override_variables parameter
        # to run_tasks (enables skipping user input from the command line)
        if name in context.variables:
            continue

        prompt = prompt_var.get("prompt") or f"Enter a value for {name}"
        private = prompt_var.get("private")
        confirm = prompt_var.get("confirm")
        allow_empty = prompt_var.get("allow_empty")
        context.variables[name] = _do_prompt(prompt, private, confirm, allow_empty)


def _file_variables(files, context: Context):
    if not files:
        return

    if not isinstance(files, list):
        files = [files]

    for file in files:
        file = context.root_directory / context.jinja_string(file)
        with file.open() as f:
            raw_vars = yaml.safe_load(f)

        for var, value in raw_vars.items():
            context.variables[var] = context.jinja_string(value)


def _extract_with(context: Context, task_args: dict) -> List[dict]:
    fileglob = task_args.pop("with_fileglob", None)

    if not util.single_truthy(fileglob, allow_zero=True):
        raise InstaterError("Must provide at most one `with_*` looping attribute")

    if fileglob:
        if not fileglob.startswith("/"):
            fileglob = str(context.root_directory) + "/" + fileglob

        return [{"item": path} for path in glob(fileglob, recursive=True)]

    return [{}]


def _load_task_item(args: dict, tags: List[str], item: dict, context: Context):
    tags = [context.jinja_string(tag, extra_vars=item) for tag in tags]

    # special handling for "include" tasks since it needs to be loaded prior to
    # actually running the task (and prior to filtering out tags)
    if "include" in args:
        replaced = {key: context.jinja_object(value, extra_vars=item) for key, value in args.items()}
        _include(context, tags, **replaced)  # type: ignore
        return

    # If a list of tags to execute is given, don't even load tasks that don't match.
    if context.tags and not any(tag in context.tags for tag in tags):
        return

    # match against a Task class by finding a key in the arguments with the name
    # (a key called 'copy', 'pacman', etc)
    task_name = next((name for name in TASKS if name in args), None)
    if task_name is None:
        raise InstaterError(f"No task matched: {args}")

    TaskClass = TASKS[task_name]
    action_args = args.pop(task_name)

    # If there aren't nested args, just a scalar arg, keep it under the task name
    # (e.g. `command: "<command here>"` doesn't have nested task arguments)
    if not isinstance(action_args, dict):
        action_args = {task_name: action_args}

    all_args = {**args, **action_args}
    replaced_args = {key: context.jinja_object(value, extra_vars=item) for key, value in all_args.items()}

    try:
        context.tasks.append(TaskClass(**replaced_args))
    except (InstaterError, TypeError) as e:
        name = args.get("name")
        error_name = f"'{name}' ({task_name})" if name else f"'{task_name}'"
        raise InstaterError(f"Error loading task {error_name}: {e}")


def _load_task(task_args: dict, tags: List[str], context: Context):
    with_items = _extract_with(context, task_args)
    for item in with_items:
        _load_task_item(task_args.copy(), tags, item, context)


def _load_tasks(task_list, context: Context, extra_tags: List[str] = None):
    if not task_list:
        return

    if not isinstance(task_list, list):
        raise InstaterError("Task definitions must be in a list")

    for task_args in task_list:
        tags = task_args.pop("tags", None) or []
        if isinstance(tags, str):
            tags = [tags]

        if extra_tags:
            tags.extend(extra_tags)

        _load_task(task_args, tags, context)


# TODO: keep track of previously included files to prevent circular dependencies
def _include(context: Context, parent_tags: List[str], include: str, tags: Union[str, List[str]] = None):
    include_file = context.root_directory / include
    if not include_file.exists():
        raise InstaterError(f"Included file does not exist: {include_file}")

    with include_file.open() as f:
        tasks = yaml.safe_load(f)

    tags = tags or []
    if isinstance(tags, str):
        tags = [tags]

    tags.extend(parent_tags)

    _load_tasks(tasks, context, tags)


def _load_context(
    setup_file,
    override_variables: dict = None,
    tags: Iterable[str] = None,
    dry_run: bool = False,
) -> Context:
    setup_file = Path(setup_file)
    context = Context(setup_file.parent, override_variables or {}, tags or (), dry_run)

    if not setup_file.exists():
        raise InstaterError(f"Setup file does not exist: {setup_file}")

    _print_start(context, setup_file)

    with setup_file.open() as f:
        setup_data = yaml.safe_load(f)

    if isinstance(setup_data, list):
        if len(setup_data) > 1:
            raise InstaterError(f"Cannot specify multiple root list items in {setup_file}")
        setup_data = setup_data[0]

    _prompt_variables(setup_data.get("vars_prompt"), context)
    _file_variables(setup_data.get("vars_files"), context)
    _load_tasks(setup_data.get("tasks"), context)

    return context


def _alert_pacman_manually_installed(context: Context):
    packages = set()

    for task in context.tasks:
        if isinstance(task, pacman.Pacman):
            for package in task.packages:
                packages.update(pacman.get_package_or_group_packages(package))

    explicitly_installed_packages = pacman.get_explicitly_installed_packages()
    manually_installed = explicitly_installed_packages - packages

    if manually_installed:
        print()
        context.print(
            ":warning: The following packages were manually installed and are not accounted for by instater",
            style="yellow bold",
        )
        for package in manually_installed:
            context.print("  - " + package, style="bold")


def run_tasks(
    setup_file,
    override_variables: dict = None,
    tags: Iterable[str] = None,
    dry_run: bool = False,
    skip_tasks: bool = False,
):
    context = _load_context(setup_file, override_variables, tags, dry_run)

    if not skip_tasks:
        for task in context.tasks:
            task.run_task(context)
            print()

    context.print_summary()

    if shutil.which("pacman"):
        _alert_pacman_manually_installed(context)

    return context



from ..context import Context
from . import Task


class Debug(Task):
    def __init__(self, debug: str, **kwargs):
        super().__init__(**kwargs)

        self.debug = debug

    def run_action(self, context: Context) -> bool:
        context.print(self.debug, style="white bold")
        return False



import re
import shutil
import subprocess
from pathlib import Path
from typing import Iterable, List, Optional, Union

from passlib.hash import sha512_crypt  # type: ignore

from .exceptions import InstaterError

Bool = Union[str, bool, int, None]


class ShellResult:
    def __init__(self, return_code: int, stdout: str, stderr: str):
        self.return_code = return_code
        self.stdout = stdout
        self.stderr = stderr


def shell(
    command: Union[str, List[str]],
    directory: Union[str, Path] = None,
    become: str = None,
    valid_return_codes: Optional[Iterable[int]] = (0,),
) -> ShellResult:
    if become:
        if isinstance(command, str):
            command = f"sudo -u {become} {command}"
        else:
            command = ["sudo", "-u", become] + command

    shell = isinstance(command, str)
    result = subprocess.run(command, cwd=directory, shell=shell, capture_output=True)
    if valid_return_codes and result.returncode not in valid_return_codes:
        error = result.stderr.decode("utf-8")
        raise InstaterError(f"Unexpected error from '{command}' (exit code {result.returncode}):\n\n{error}")

    return ShellResult(result.returncode, result.stdout.decode("utf-8").strip(), result.stderr.decode("utf-8").strip())


def update_file_metadata(
    path: Path,
    owner: Optional[str],
    group: Optional[str],
    mode: Optional[int],
    dry_run: bool,
) -> bool:
    updated = False

    # If running with dry_run and the path does not exist, the below checks error
    if dry_run and not path.exists():
        return True

    if (owner or group) and (path.owner() != owner or path.group() != group):
        if not dry_run:
            shutil.chown(path, owner, group)
        updated = True

    current_mode = path.stat().st_mode & 0o777
    if mode is not None and (current_mode != mode):
        if not dry_run:
            path.chmod(mode)
        updated = True

    return updated


def boolean(input: Bool) -> bool:
    # converting to string covers int (0) and bool (False) cases
    return bool(input) and str(input).lower() not in ("no", "false", "0")


def password_hash(password: str, hashtype: str = "sha512") -> str:
    if hashtype != "sha512":
        raise InstaterError(f"password_hash hashtype must be sha512, found '{hashtype}'")

    return sha512_crypt.hash(password)


_SNAKE_CASE = re.compile(r"(?<!^)(?=[A-Z])")


def snake_case(camel_case: str) -> str:
    return _SNAKE_CASE.sub("_", camel_case).lower()


def single_truthy(*args, allow_zero: bool = False) -> bool:
    count = tuple(map(bool, args)).count(True)
    if allow_zero:
        return count <= 1
    else:
        return count == 1



import time

from ..context import Context
from ..exceptions import InstaterError
from ..util import snake_case

TASKS = {}


class Task:
    def __init__(
        self,
        name: str = None,
        when: str = None,
        register: str = None,
    ):
        self.name = name or "Unnamed " + snake_case(type(self).__name__)
        self.when = when
        self.register = register

    def __init_subclass__(cls):
        TASKS[snake_case(cls.__name__)] = cls

    def run_task(self, context: Context) -> bool:
        context.print(f"TASK [{self.name}]", style="black bold on blue", justify="left")

        start = time.time()

        if self.when and context.jinja_string("{{ (" + self.when + ") | bool }}") == "False":
            changed = False
        else:
            changed = self.run_action(context)

        duration = context.duration(start)
        if changed:
            context.statuses["changed"] += 1
            context.print(f"changed {duration}", style="yellow bold")
        else:
            context.statuses["skipped"] += 1
            context.print(f"skipped {duration}", style="blue")

        if self.register:
            if self.register in context.variables:
                raise InstaterError(f"Task registered as '{self.register}' conflicts with an existing variable")
            context.variables[self.register] = {"changed": changed}

        return changed

    def run_action(self, context: Context) -> bool:
        raise NotImplementedError



from pathlib import Path

from setuptools import find_packages, setup


ROOT_DIRECTORY = Path(__file__).resolve().parent

description = "An easy solution for system/dotfile configuration"
readme = (ROOT_DIRECTORY / "README.md").read_text()
changelog = (ROOT_DIRECTORY / "CHANGELOG.md").read_text()
long_description = readme + "\n\n" + changelog

version = (ROOT_DIRECTORY / "instater" / "VERSION").read_text().strip()


DEV_REQUIRES = [
    "black==22.3.0",
    "coverage==6.0.2",
    "flake8==4.0.1",
    "flake8-bugbear==21.9.2",
    "isort==5.9.3",
    "mypy==0.910",
    "pytest==6.2.5",
    "pytest-cov==3.0.0",
    "twine==3.4.2",
]

setup(
    name="instater",
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    author="Naya Verdier",
    url="https://github.com/nayaverdier/instater",
    license="MIT",
    packages=find_packages(exclude=("tests",)),
    entry_points={"console_scripts": ["instater = instater.cli:main"]},
    install_requires=[
        "Jinja2==3.0.2",
        "PyYAML==6.0",
        "passlib==1.7.4",
        "rich==10.12.0",
    ],
    python_requires=">=3.7",
    extras_require={
        "dev": DEV_REQUIRES,
    },
    include_package_data=True,
)



from typing import List, Union

from .. import util
from ..context import Context
from . import Task


class Command(Task):
    def __init__(
        self,
        command: Union[str, List[str]],
        condition: str = None,
        condition_code: int = 0,
        become: str = None,
        directory: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if isinstance(command, str):
            command = [command]

        self.commands = command
        self.condition = condition
        self.condition_code = int(condition_code)
        self.become = become
        self.directory = directory

    def run_action(self, context: Context) -> bool:
        if self.condition:
            # TODO: only consider specific return codes as valid?
            result = util.shell(self.condition, self.directory, valid_return_codes=None)
            if result.return_code != self.condition_code:
                return False

        if not context.dry_run:
            for command in self.commands:
                util.shell(command, self.directory, become=self.become)

        return True



from pathlib import Path

from instater.exceptions import InstaterError

from .. import util
from ..context import Context
from . import Task


class Git(Task):
    def __init__(
        self, *, repo: str, dest: str, depth: int = None, fetch_tags: bool = True, become: str = None, **kwargs
    ):
        super().__init__(**kwargs)
        self.repo = repo
        self.dest = Path(dest)
        self.depth = str(depth)
        self.tags_flag = "--tags" if fetch_tags else "--no-tags"
        self.become = become

    def _clone(self):
        command = ["git", "clone", self.repo]
        if self.depth is not None:
            command += ["--depth", self.depth]

        command.append(str(self.dest))
        util.shell(command, become=self.become)

    def _get_remote(self) -> str:
        result = util.shell(["git", "config", "--get", "remote.origin.url"], directory=self.dest, become=self.become)
        return result.stdout

    def _should_pull(self) -> bool:
        result = util.shell(["git", "fetch", "--dry-run", self.tags_flag], directory=self.dest, become=self.become)
        return result.stdout != "" or result.stderr != ""

    def _pull(self):
        branch = util.shell(["git", "branch", "--show-current"], directory=self.dest, become=self.become).stdout
        util.shell(["git", "pull", "origin", branch, self.tags_flag], directory=self.dest, become=self.become)

    def run_action(self, context: Context):
        if not self.dest.exists():
            if not context.dry_run:
                self._clone()
            return True

        if not (self.dest / ".git").exists():
            raise InstaterError(f"Git destination directory exists, but is not a git repo: {self.dest}")

        if self._get_remote() != self.repo:
            raise InstaterError(f"Git remote does not match current local git repo: {self.dest}")

        if self._should_pull():
            if not context.dry_run:
                self._pull()
            return True

        return False



import pkgutil
import sys
from pathlib import Path


def _import_exception_handler(module_name: str):
    print(f"Exception when importing module {module_name}")
    raise


_loaded = set()


def import_submodules(module_name: str):
    if module_name in _loaded:
        return

    module = sys.modules[module_name]

    path = Path(module.__file__).parent.absolute()
    prefix = module.__name__
    paths = [str(path)]

    for loader, submodule_name, is_pkg in pkgutil.walk_packages(paths, prefix + ".", onerror=_import_exception_handler):
        if is_pkg or submodule_name.split(".")[-1].startswith("_"):
            continue

        loader.find_module(submodule_name).load_module(submodule_name)  # type: ignore

    _loaded.add(module_name)
